# -*- coding: utf-8 -*-

import os
import io
import glob

# concatenate a set of files Web of Science data files
# files are named filebas*.txt in the folder datapath
# output to outfile
#
def mergeWOSFiles(datapath, filebase, outfile):
    files = [f for f in glob.glob(os.path.join(datapath, filebase+"*.txt"))]

    of = io.open(os.path.join(datapath, outfile), 'w', encoding='utf-8')

    # concatenate files, ditch header from all files except the first one
    cnt = 0
    for idx, fname in enumerate(files):
        print("Processing file %d %s"%(idx, fname))
        df = io.open(fname, 'r', encoding='utf-8')
        for lnum, li in enumerate(df):
            if idx == 0 or lnum > 0:
                of.writelines([li[0:131000]])
                cnt += 1
        print("%d lines"%cnt)
        df.close()
    of.close()
    print("%d lines"%cnt)




# -*- coding: utf-8 -*-
"""
Created on Thu Aug 14 10:19:52 2014

@author: rich

add intercluster fraction, cluster diversity, bridging and centrality to each node

"""

import numpy as np


def basicClusteringProperties(network, clustering):
    """
    compute diversity and related properties for the given clustering
    adds results to node attributes
    """
    if clustering == 'Cluster':
        properties = ['InterclusterFraction', 'ClusterDiversity', 'ClusterBridging', 'ClusterCentrality']
    else:
        properties = ['fracIntergroup_' + clustering, 'diversity_' + clustering, 'bridging_' + clustering,
                      'centrality_' + clustering]
    results = {prop: {} for prop in properties}

    clusters = {}
    # iterate over each node
    for node in network:
        nodedata = network.nodes[node]
        cluster = nodedata[clustering] if clustering in nodedata else None
        if cluster is not None and cluster != '':
            # build list of nodes in each cluster
            if cluster not in clusters:
                clusters[cluster] = []
            clusters[cluster].append(node)
            clusterCounts = {}  # dict of cluster name and count
            nIntergroup = 0
            degree = float(network.degree(node))

            # walk neighbors and save cluster info
            for neighbor in network.neighbors(node):
                neighborCluster = network.nodes[neighbor][clustering]
                if neighborCluster not in clusterCounts:
                    clusterCounts[neighborCluster] = 0.0
                clusterCounts[neighborCluster] += 1.0
                if neighborCluster is not None and cluster != neighborCluster:
                    nIntergroup += 1

            # compute diversity and related properties
            nGroups = len(clusterCounts)
            fracIntergroup = float(nIntergroup) / degree if (degree > 0) else 0
            diversity = 0
            p = np.array(list(clusterCounts.values())) / degree
            p = p[np.nonzero(p)]
            diversity = -np.sum(p * np.log(p))
            bridging = 0 if nGroups < 2 else diversity * float(nIntergroup) / (nGroups - 1)
            centrality = (1 - fracIntergroup) * degree / (1 + diversity)

            results[properties[0]][node] = fracIntergroup
            results[properties[1]][node] = diversity
            results[properties[2]][node] = bridging
            results[properties[3]][node] = centrality
    # normalize values within each cluster
    for nodes in clusters.values():
        def normalize(results, nodes, attr):
            def normalizeArray(arr):
                npArr = np.array(arr)
                sd = np.std(npArr)
                if sd != 0:
                    mn = np.mean(npArr)
                    return ((npArr - mn) / sd).tolist()
                return arr

            vals = normalizeArray([results[attr][node] for node in nodes])
            for v in zip(nodes, vals):
                results[attr][v[0]] = v[1]

        normalize(results, nodes, properties[1])
        normalize(results, nodes, properties[2])
        normalize(results, nodes, properties[3])

    return results



# -*- coding: utf-8 -*-
#
#
# Build network from similarity matrix by thresholding
# or from matrix of possibly weighted connections 
#
# Add data to nodes - cluster, netowrk proerties, layout coordinates
#

import numpy as np
import pandas as pd
import math
from collections import Counter
from scipy.sparse import dok_matrix
from scipy.sparse import csr_matrix
import networkx as nx

'''
import Network.ClusteringProperties as cp
from Network.DrawNetwork import draw_network_categorical
#from InteractiveNetworkViz import drawInteractiveNW
from Network.louvain import generate_dendrogram
from Network.louvain import partition_at_level
from Network.tSNELayout import runTSNELayout, runUMAPlayout
from Network.ClusterLayout import run_cluster_layout
'''

import ClusteringProperties as cp
from DrawNetwork import draw_network_categorical
from louvain import generate_dendrogram
from louvain import partition_at_level
from tSNELayout import runTSNELayout, runUMAPlayout
from ClusterLayout import run_cluster_layout


# build sparse feature matrix with optional idf weighting
# each row is a document, each column is a tag
# weighting assumes each term occurs once in each doc it appears in
def buildFeatures(df, tagHist, idf, tagAttr):
    allTags = tagHist.keys()
    # build tag-index mapping
    tagIdx = dict(zip(allTags, range(len(allTags))))
    # build feature matrix
    print("Build feature matrix")
    nDoc = len(df)
    features = dok_matrix((nDoc, len(tagIdx)), dtype=float)
    row = 0
    for tagList in df[tagAttr]:
        tagList = [k for k in tagList if k in tagIdx]
        if len(tagList) > 0:
            for tag in tagList:
                if idf:
                    docFreq = tagHist[tag]
                    features[row, tagIdx[tag]] = math.log(nDoc/float(docFreq), 2.0)
                else:
                    features[row, tagIdx[tag]] = 1.0
        else:
            print("Document with no tags (%s)" % tagAttr)
        row += 1
    return csr_matrix(features)


# compute cosine similarity
# f is a (sparse) feature matrix
def simCosine(f):
    # compute feature matrix dot product
    fdot = np.array(f.dot(f.T).todense())
    # get inverse feature vector magnitudes
    invMag = np.sqrt(np.divide(1.0, np.diag(fdot)))
    # set NaN to zero
    invMag[np.isinf(invMag)] = 0
    # get cosine sim by elementwise multiply by inverse magnitudes
    sim = (fdot * invMag).T * invMag
    return sim


def threshold(sim, linksPer=4, connect_isolated_pairs=True):
    '''
    threshold similarity matrix - threshold by minimum maximum similarity of each node. Then if there 
    are too many links, thin links of each node propoertional to their abundance, but leave at least 
    one link from each node
    Parameters:
        sim - the similarity matrix
        linksPer - target connectivity
        connect_isolated_pairs - if true, connect isolated reciprocal pairs of nodes to their next 
        most similar neighbors
    '''
    simvals = sim.copy()
    nnodes = sim.shape[0]
    targetL = nnodes*linksPer
    # threshold on minimum max similarity in each row, this keeps at least one link per row
    mxsim = sim.max(axis=0)
    thr = mxsim[mxsim > 0].min()
    sim[sim < thr] = 0.0
    nL = (sim > 0).sum()
    # if too many links, keep equal fraction of links in each row,
    # minimally 1 per row, keep highest similarity links
    if nL > targetL:
        # get fraction to keep
        frac = targetL/float(nL)
        # sort the rows
        indices = np.argsort(sim, axis=1)
        # get number of elements in each row
        nonzero = np.round(np.maximum((frac*((sim > 0).sum(axis=1))), 1)).astype(int)
        # get minimum index to keep in each row
        minelement = (nnodes - nonzero).astype(int)
        # in each row, set values below number to keep to zero
        for i in range(nnodes):
            sim[i][indices[i][:minelement[i]]] = 0.0
        if connect_isolated_pairs:
            # for isolated reciprocal pairs, keep next lower similarity link
            # fisrt find all reciprocal pairs
            upper = np.triu(sim)
            recip = np.argwhere((upper > 0) & (np.isclose(upper, np.tril(sim).T, 1e-14)))
            # get isolated reciprocal pairs
            links = sim > 0
            isolated = (links[recip[:, 0]].sum(axis=1) == 1) & (links[recip[:, 1]].sum(axis=1) == 1) 
            # get all nodes involved in isolated pairs
            isolated_recip = recip[isolated].flatten()
            # add next most similar link        
            sim[isolated_recip, indices[isolated_recip, -2]] = simvals[isolated_recip, indices[isolated_recip, -2]]
    return sim


# build cluster name based on keywords that occur commonly in the cluster
# if wtd, then weigh keywords based on local frequency relative to global freq
# NOTE TO RICH:  I ADDED HERE PARAMETERS TO NAME THE CLUSTER NAME AND TOP TAGS COLUMNS
def buildClusterNames(df, allTagHist, tagAttr, 
                      clAttr='Cluster', clusterName='cluster_name', 
                      topTags='top_tags', wtd=True):
    allVals = np.array(list(allTagHist.values()), dtype=float)
    allFreq = dict(zip(allTagHist.keys(), allVals/allVals.sum()))
    #clusters = df['clusId'].unique()
    clusters = df[clAttr].unique()
    df[clusterName] = ''
    df[topTags] = ''
    clusInfo = []
    for clus in clusters:
        clusRows = df[clAttr] == clus
        nRows = clusRows.sum()
        if nRows > 0:
            tagHist = Counter([k for tagList in df[tagAttr][clusRows] for k in tagList if k in allTagHist])
            if wtd:
                vals = np.array(list(tagHist.values()), dtype=float)
                freq = dict(zip(tagHist.keys(), vals/vals.sum()))
                # weight tags, only include tag if it is more common than global
                wtdTag = [(item[0], item[1]*math.sqrt(freq[item[0]]/allFreq[item[0]]))
                          for item in tagHist.most_common() if freq[item[0]] > allFreq[item[0]]]
                wtdTag.sort(key=lambda x: x[1], reverse=True)
                topTag = [item[0] for item in wtdTag[:10]]
            else:
                topTag = [item[0] for item in tagHist.most_common()][:10]
            # remove unigrams that make up n-grams with n > 1
            topSet = set(topTag)
            removeTag = set()
            for tag in topTag:
                tags = tag.split(' ')
                if len(tags) > 1 and all(k in topSet for k in tags):
                    removeTag.update(tags)
            topTag = [k for k in topTag if k not in removeTag]
            # build and store name
            clName = ', '.join(topTag[:5])
            df.loc[clusRows, clusterName] = clName
            df.loc[clusRows, topTags] = ', '.join(topTag)
            clusInfo.append((clus, nRows, clName))
    df[topTags] = df[topTags].str.split(',')
    clusInfo.sort(key=lambda x: x[1], reverse=True)
    for info in clusInfo:
        print("Cluster %s, %d nodes, name: %s" % info)


def buildNetworkFromNodesAndEdges(nodesdf, edgedf, outname=None,
                                  nodesname=None, edgesname=None, plotfile=None,
                                  doLayout=True, clusteredLayout=False,
                                  tagHist=None, tagAttr=None):
    # add clusters and attributes
    nw = buildNetworkX(edgedf)
    addLouvainClusters(nodesdf, nw=nw)
    addNetworkAttributes(nodesdf, nw=nw)
    if tagHist and tagAttr:
        buildClusterNames(nodesdf, tagHist, tagAttr)
    if doLayout:
        add_layout(nodesdf, nw=nw, clustered=clusteredLayout)
        if plotfile is not None:
            #drawInteractiveNW(df, nw=nw, plotfile=plotfile)
            draw_network_categorical(nw, nodesdf, plotfile=plotfile, draw_edges=False)
    # output to csv
    if nodesname is not None and edgesname is not None:
        print("Writing nodes and edges to files")
        nodesdf.to_csv(nodesname, index=False)
        edgedf.to_csv(edgesname, index=False)
    # output to xlsx
    if outname is not None:
        print("Writing network to file")
        writer = pd.ExcelWriter(outname)
        nodesdf.to_excel(writer, 'Nodes', index=False)
        edgedf.to_excel(writer, 'Links', index=False)
        writer.save()
    return nodesdf, edgedf


# build network helper function
# thresholds similarity, computes clusters and other attributes, names clusters if applicable
# draws network, saves plot and data to files
# return nodesdf, edgedf
def _buildNetworkHelper(df, sim, linksPer=4, outname=None,
                        nodesname=None, edgesname=None, plotfile=None,
                        doLayout=True, clusteredLayout=False, tagHist=None, tagAttr=None):
    # threshold
    if linksPer > 0:
        print("Threshold similarity")
        sim = threshold(sim, linksPer=linksPer)
    # make edge dataframe
    edgedf = matrixToLinkDataFrame(sim)

    return buildNetworkFromNodesAndEdges(df, edgedf, outname=outname,
                                         nodesname=nodesname, edgesname=edgesname, plotfile=plotfile,
                                         doLayout=doLayout, clusteredLayout=clusteredLayout,
                                         tagHist=tagHist, tagAttr=tagAttr)


# build network, linking based on common tags, tag lists in column named tagAttr
# color_attr - the attribute to color the nodes by
# outname - name of xlsx file to output network to
# nodesname - name of file for nodes csv
# edgesname - name of file for edge csv
# plotfile - name of file for plot image
# doLayout - if true, run layout
# draw - if True and if running layout, then draw the network and possibly save image to file (if plotfile is given)
# return nodesdf, edgedf
def buildTagNetwork(df, color_attr="Cluster", tagAttr='eKwds', dropCols=[], outname=None,
                    nodesname=None, edgesname=None, plotfile=None, idf=True,
                    toFile=True, doLayout=True, clusteredLayout=False, linksPer=4, minTags=0):
    print("Building document network")
    df = df.copy()  # so passed-in dataframe is not altered
    # make histogram of tag frequencies, only include tags with > 1 occurence
    tagHist = dict([item for item in Counter([k for kwList in df[tagAttr] 
                                              for k in kwList]).most_common() if item[1] > 1])
    # filter tags to only include 'active' tags - tags which occur twice or more in the doc set
    df[tagAttr] = df[tagAttr].apply(lambda x: [k for k in x if k in tagHist])
    # filter docs to only include docs with a minimum number of 'active' tags
    df = df[df[tagAttr].apply(lambda x: len(x) >= minTags)]
    # build document-keywords feature matrix
    features = buildFeatures(df, tagHist, idf, tagAttr)
    # compute similarity
    print("Compute similarity")
    sim = simCosine(features)
    # avoid self-links
    np.fill_diagonal(sim, 0)
    df['id'] = range(len(df))
    df.drop(dropCols, axis=1, inplace=True)
    return _buildNetworkHelper(df, sim, outname=outname,
                               nodesname=nodesname, edgesname=edgesname, plotfile=plotfile,
                               doLayout=doLayout, clusteredLayout=clusteredLayout,
                               tagHist=tagHist, tagAttr=tagAttr, linksPer=linksPer)


# build network given node dataframe and similarity matrix
# color_attr is the attribute to color the nodes by
# outname - name of xlsx file to output network to
# nodesname - name of file for nodes csv
# edgesname - name of file for edge csv
# plotfile - name of file for plot image
# doLayout - if true, run layout
# draw - if True and if running layout, then draw the network and possibly save image to file (if plotfile is given)
def buildSimilarityNetwork(df, sim, color_attr="Cluster", outname=None,
                           nodesname=None, edgesname=None, plotfile=None,
                           toFile=True, doLayout=True, linksPer=4):
    df['id'] = range(len(df))
    return _buildNetworkHelper(df, sim, outname=outname,
                               nodesname=nodesname, edgesname=edgesname, plotfile=plotfile,
                               toFile=toFile, doLayout=doLayout, linksPer=linksPer)


# build link dataframe from matrix where non-zero element is a link
def matrixToLinkDataFrame(mat, undirected=False, include_weights=True):
    if undirected:  # make symmetric then take upper triangle
        mat = np.triu(np.maximum(mat, mat.T))   
    links = np.transpose(np.nonzero(mat))
    linkList = [{'Source': l[0], 'Target': l[1]} for l in links]
    if include_weights:
        linkList = [{'Source': l[0], 'Target': l[1], 'weight': mat[l[0], l[1]]} for l in links]
    else:
        linkList = [{'Source': l[0], 'Target': l[1]} for l in links]
    return pd.DataFrame(linkList)


def buildNetworkX(linksdf, id1='Source', id2='Target', directed=False):
    linkdata = [(getattr(link, id1), getattr(link, id2)) for link in linksdf.itertuples()]
    g = nx.DiGraph() if directed else nx.Graph()
    g.add_edges_from(linkdata)
    return g


# add a computed network attribute to the node attribute table
#
def _add_attr(nodesdf, attr, vals):
    nodesdf[attr] = nodesdf['id'].map(vals)


# add network structural attributes to nodesdf
# clusVar is the attribute to use for computing bridging etc
def addNetworkAttributes(nodesdf, linksdf=None, nw=None, groupVars=["Cluster"], isDirected=False):
    if nw is None:
        nw = buildNetworkX(linksdf)
    if isDirected:
        _add_attr(nodesdf, "InDegree", dict(nw.in_degree()))
        _add_attr(nodesdf, "OutDegree", dict(nw.out_degree()))
    else:
        _add_attr(nodesdf, "Degree", dict(nw.degree()))

    # add bridging, cluster centrality etc. for one or more grouping variables
    for groupVar in groupVars:
        if len(nx.get_node_attributes(nw, groupVar)) == 0:
            vals = {k: v for k, v in dict(zip(nodesdf['id'], nodesdf[groupVar])).items() if k in nw}
            nx.set_node_attributes(nw, vals, groupVar)
        grpprop = cp.basicClusteringProperties(nw, groupVar)
        for prop, vals in grpprop.items():
            _add_attr(nodesdf, prop, vals)


# compute and add Louvain clusters to node dataframe
def addLouvainClusters(nodesdf, linksdf=None, nw=None, clusterLevel=0):
    def mergePartitionData(g, p, name):
        return {node: (name + '_' + str(p[node]) if node in p else None) for node in g.nodes()}

    def getPartitioning(i, g, dendo, clusterings):
        p = partition_at_level(dendo, len(dendo) - 1 - i)
        clustering = "Cluster"
        vals = mergePartitionData(g, p, clustering)
        clusterings[clustering] = vals

    if nw is None:
        nw = buildNetworkX(linksdf)
    print("Computing clustering")
    if isinstance(nw, nx.DiGraph):
        gg = nx.Graph(nw)
    else:
        gg = nw
    clusterings = {}
    dendo = generate_dendrogram(gg)
    #    for i in range(len(dendo)):
    #        getPartitioning(i, g, dendo, clusterings)
    depth = min(clusterLevel, len(dendo) - 1)
    getPartitioning(depth, gg, dendo, clusterings)
    # add cluster attr to dataframe
    for grp, vals in clusterings.items():
        _add_attr(nodesdf, grp, vals)
        nodesdf[grp].fillna('No Cluster', inplace=True)


def add_layout(nodesdf, linksdf=None, nw=None, clustered=True, cluster='Cluster'):
    print("Running graph layout")
    if nw is None:
        nw = buildNetworkX(linksdf)
    if clustered:
        #layout, _ = runTSNELayout(nw, nodesdf=nodesdf, cluster='Cluster')
        #layout, _ = runUMAPlayout(nw, nodesdf=nodesdf, cluster='Cluster')
        layout, _ = run_cluster_layout(nw, nodes_df=nodesdf, cluster=cluster)
    else:
        layout, _ = runTSNELayout(nw, nodesdf=nodesdf)
        # layout, _ = runUMAPlayout(nw)
    nodesdf['x'] = nodesdf['id'].apply(lambda x: layout[x][0] if x in layout else 0.0)
    nodesdf['y'] = nodesdf['id'].apply(lambda x: layout[x][1] if x in layout else 0.0)
    return layout


def add_force_directed_layout(nodesdf, linksdf=None, nw=None, iterations=100):
    print("Running force-directed graph layout")
    if nw is None:
        nw = buildNetworkX(linksdf)
    layout = nx.spring_layout(nw, iterations=iterations)
    nodesdf['x_force_directed'] = nodesdf['id'].apply(lambda x: layout[x][0] if x in layout else 0.0)
    nodesdf['y_force_directed'] = nodesdf['id'].apply(lambda x: layout[x][1] if x in layout else 0.0)
    return layout



# -*- coding: utf-8 -*-

import math
import networkx as nx

from bokeh.io import show, output_file, output_notebook
from bokeh.plotting import figure
from bokeh.models import GraphRenderer, StaticLayoutProvider, Circle, MultiLine
from bokeh.models import HoverTool, TapTool, BoxSelectTool
from bokeh.models import NodesAndLinkedEdges
from bokeh.core.properties import field

#from bokeh.palettes import Spectral8

import Network.DrawNetwork as dn

# draw network using categorical attribute for node colors
# draw using Bokeh
# nodes (df) must include x,y coordinates
# links are specified either in networkx object (nw) or list of links in a DataFrame
#
def drawInteractiveNW(df, nw=None, edgesdf=None, color_attr="Cluster",
                      label_attr="name", title="Interactive Network Visualization",
                      plotfile=None, inline=False):
    def buildNetworkX(linksdf, id1='Source', id2='Target', directed=False):
        linkdata = [(getattr(link, id1), getattr(link, id2)) for link in linksdf.itertuples()]
        g = nx.DiGraph() if directed else nx.Graph()
        g.add_edges_from(linkdata)
        return g

    if inline:
        output_notebook()
    if plotfile:
        output_file(plotfile+".html")
    if nw is None:
        if edgesdf is None:
            print("Must specify either network or edges DataFrame")
            return
        nw = buildNetworkX(edgesdf)
    node_colors, edge_colors, attr_colors = dn.getCategoricalColors(nw, df, color_attr, None, True, None)
    xmin = df['x'].min()
    xmax = df['x'].max()
    ymin = df['y'].min()
    ymax = df['y'].max()
    rng = max((xmax-xmin), (ymax-ymin))
    nNodes = len(df)
    size = 4*rng/math.sqrt(nNodes)
    node_indices = list(range(nNodes))

    tooltips=[
        (label_attr, "@"+label_attr),
        (color_attr, "@"+color_attr)
    ]

    plot = figure(title=title, plot_width=800, plot_height=800, x_range=(xmin-size, xmax+size),
                  y_range=(ymin-size, ymax+size),
                  tools="pan,wheel_zoom,box_zoom,reset", toolbar_location="right", output_backend="webgl")
    plot.add_tools(HoverTool(tooltips=tooltips), TapTool(), BoxSelectTool())

    graph = GraphRenderer()

    # set node renderer and data
    graph.node_renderer.glyph = Circle(size=size, fill_color="fill_color", line_color='gray')
    graph.node_renderer.selection_glyph = Circle(size=size, fill_color="fill_color", line_color='black', line_width=2)
    graph.node_renderer.hover_glyph = Circle(size=size, fill_color="fill_color", line_color='black')
    graph.node_renderer.data_source.data = {'index': node_indices,
                                            label_attr: df[label_attr].tolist(),
                                            'fill_color': node_colors,
                                            color_attr: df[color_attr].fillna('').tolist(),
                                            'x': df['x'].tolist(),
                                            'y': df['y'].tolist(),
                                            }
    # set edge renderer and data
    graph.edge_renderer.glyph = MultiLine(line_color="line_color", line_width=1)
    graph.edge_renderer.data_source.data = {'start': [e[0] for e in nw.edges],
                                            'end': [e[1] for e in nw.edges],
                                            'line_color': edge_colors
                                            }
    # set layout
    graph_layout = dict(zip(node_indices, zip(df['x'], df['y'])))
    graph.layout_provider = StaticLayoutProvider(graph_layout=graph_layout)

    # set legend by adding dummy glyph with legend
    plot.circle( x='x', y='y', radius=0.0, color='fill_color', legend=field(color_attr), source=graph.node_renderer.data_source.data)
    plot.legend.location = "top_left"
    plot.legend.padding = 0
    plot.legend.margin = 0

    graph.selection_policy = NodesAndLinkedEdges()

    # hide axes and grids
    plot.xaxis.visible = False
    plot.yaxis.visible = False
    plot.xgrid.visible = False
    plot.ygrid.visible = False

    plot.renderers.append(graph)

    show(plot)



# -*- coding: utf-8 -*-
"""
Created on Thu Jun 30 14:41:04 2016

@author: rich williams

Modified and extended networkx drawing code

#    Copyright (C) 2004-2016 by
#    Aric Hagberg <hagberg@lanl.gov>
#    Dan Schult <dschult@colgate.edu>
#    Pieter Swart <swart@lanl.gov>
#    All rights reserved.
#    BSD license.
#
# Author: Aric Hagberg (hagberg@lanl.gov)

**********
Matplotlib
**********
Draw networks with matplotlib.
See Also
--------
matplotlib:     http://matplotlib.org/
pygraphviz:     http://pygraphviz.github.io/
"""

import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import matplotlib.cbook as cb
from matplotlib.colors import colorConverter, Colormap
from matplotlib.collections import LineCollection
from matplotlib.collections import PolyCollection
import math
import numbers
import numpy as np
import networkx as nx
from networkx.drawing.nx_pylab import draw_networkx_labels

__all__ = ['_draw_networkx_']


def _draw_networkx_(G, pos=None, with_labels=True, draw_edges=True, **kwds):
    """Draw the graph G using Matplotlib.

    Draw the graph with Matplotlib with options for node positions,
    labeling, titles, and many other drawing features.
    See draw() for simple drawing without labels or axes.

    Parameters
    ----------
    G : graph
       A networkx graph
    pos : dictionary, optional
       A dictionary with nodes as keys and positions as values.
       If not specified a spring layout positioning will be computed.
       See networkx.layout for functions that compute node positions.
    with_labels :  bool, optional (default=True)
       Set to True to draw labels on the nodes.
    ax : Matplotlib Axes object, optional
       Draw the graph in the specified Matplotlib axes.
    nodelist : list, optional (default G.nodes())
       Draw only specified nodes
    edgelist : list, optional (default=G.edges())
       Draw only specified edges
    node_size : scalar or array, optional (default=300)
       Size of nodes.  If an array is specified it must be the
       same length as nodelist.
    node_color : color string, or array of floats, (default='r')
       Node color. Can be a single color format string,
       or a  sequence of colors with the same length as nodelist.
       If numeric values are specified they will be mapped to
       colors using the cmap and vmin,vmax parameters.  See
       matplotlib.scatter for more details.
    node_shape :  string, optional (default='o')
       The shape of the node.  Specification is as matplotlib.scatter
       marker, one of 'so^>v<dph8'.
    highlight : array of boolean, if true a black ring is draw round the node
    alpha : float, optional (default=1.0)
       The node and edge transparency
    cmap : Matplotlib colormap, optional (default=None)
       Colormap for mapping intensities of nodes
    vmin,vmax : float, optional (default=None)
       Minimum and maximum for node colormap scaling
    linewidths : [None | scalar | sequence]
       Line width of symbol border (default =1.0)
    width : float, optional (default=1.0)
       Line width of edges
    edge_color : color string, or array of floats (default='r')
       Edge color. Can be a single color format string,
       or a sequence of colors with the same length as edgelist.
       If numeric values are specified they will be mapped to
       colors using the edge_cmap and edge_vmin,edge_vmax parameters.
    edge_cmap : Matplotlib colormap, optional (default=None)
       Colormap for mapping intensities of edges
    edge_vmin,edge_vmax : floats, optional (default=None)
       Minimum and maximum for edge colormap scaling
    style : string, optional (default='solid')
       Edge line style (solid|dashed|dotted,dashdot)
    labels : dictionary, optional (default=None)
       Node labels in a dictionary keyed by node of text labels
    font_size : int, optional (default=12)
       Font size for text labels
    font_color : string, optional (default='k' black)
       Font color string
    font_weight : string, optional (default='normal')
       Font weight
    font_family : string, optional (default='sans-serif')
       Font family
    label : string, optional
        Label for graph legend
    Examples
    --------
    >>> G=nx.dodecahedral_graph()
    >>> nx.draw(G)
    >>> nx.draw(G,pos=nx.spring_layout(G)) # use spring layout
    >>> import matplotlib.pyplot as plt
    >>> limits=plt.axis('off') # turn of axis
    Also see the NetworkX drawing examples at
    http://networkx.github.io/documentation/latest/gallery.html
    See Also
    --------
    draw()
    draw_networkx_nodes()
    draw_networkx_edges()
    draw_networkx_labels()
    draw_networkx_edge_labels()
    """
    if pos is None:
        pos = nx.drawing.spring_layout(G)  # default to spring layout

    draw_nx_nodes(G, pos, **kwds)
    if draw_edges:
        draw_nx_tapered_edges(G, pos, **kwds)
    if with_labels:
        draw_networkx_labels(G, pos, **kwds)
    plt.draw_if_interactive()


def draw_nx_tapered_edges(G, pos,
                          edgelist=None,
                          width=0.5,
                          edge_color='k',
                          style='solid',
                          alpha=1.0,
                          edge_cmap=None,
                          edge_vmin=None,
                          edge_vmax=None,
                          ax=None,
                          label=None,
                          highlight=None,
                          tapered=False,
                          **kwds):
    """Draw the edges of the graph G.

    This draws only the edges of the graph G.

    Parameters
    ----------
    G : graph
       A networkx graph
    pos : dictionary
       A dictionary with nodes as keys and positions as values.
       Positions should be sequences of length 2.
    edgelist : collection of edge tuples
       Draw only specified edges(default=G.edges())
    width : float, or array of floats
       Line width of edges (default=1.0)
    edge_color : color string, or array of floats
       Edge color. Can be a single color format string (default='r'),
       or a sequence of colors with the same length as edgelist.
       If numeric values are specified they will be mapped to
       colors using the edge_cmap and edge_vmin,edge_vmax parameters.
    style : string
       Edge line style (default='solid') (solid|dashed|dotted,dashdot)
    alpha : float
       The edge transparency (default=1.0)
    edge_ cmap : Matplotlib colormap
       Colormap for mapping intensities of edges (default=None)
    edge_vmin,edge_vmax : floats
       Minimum and maximum for edge colormap scaling (default=None)
    ax : Matplotlib Axes object, optional
       Draw the graph in the specified Matplotlib axes.
    label : [None| string]
       Label for legend
    Returns
    -------
    matplotlib.collection.LineCollection
        `LineCollection` of the edges
    Examples
    --------
    >>> G=nx.dodecahedral_graph()
    >>> edges=nx.draw_networkx_edges(G,pos=nx.spring_layout(G))
    Also see the NetworkX drawing examples at
    http://networkx.github.io/documentation/latest/gallery.html
    See Also
    --------
    draw()
    draw_networkx()
    draw_networkx_nodes()
    draw_networkx_labels()
    draw_networkx_edge_labels()
    """
    if ax is None:
        ax = plt.gca()

    if edgelist is None:
        edgelist = list(G.edges())

    if not edgelist or len(edgelist) == 0:  # no edges!
        return None

    if highlight is not None and (isinstance(edge_color, str) or not cb.iterable(edge_color)):
        idMap = {}
        nodes = G.nodes()
        for i in range(len(nodes)):
            idMap[nodes[i]] = i
        ecol = [edge_color]*len(edgelist)
        eHighlight = [highlight[idMap[edge[0]]] or highlight[idMap[edge[1]]] for edge in edgelist]
        for i in range(len(eHighlight)):
            if eHighlight[i]:
                ecol[i] = '0.0'
        edge_color = ecol

    # set edge positions
    if not np.iterable(width):
        lw = np.full(len(edgelist), width)
    else:
        lw = width

    edge_pos = []
    wdScale = 0.01
    for i in range(len(edgelist)):
        e = edgelist[i]
        w = wdScale*lw[i]/2
        p0 = pos[e[0]]
        p1 = pos[e[1]]
        dx = p1[0] - p0[0]
        dy = p1[1] - p0[1]
        l = math.sqrt(dx*dx+dy*dy)
        edge_pos.append(((p0[0]+w*dy/l, p0[1]-w*dx/l), (p0[0]-w*dy/l, p0[1]+w*dx/l), (p1[0], p1[1])))

    edge_vertices = np.asarray(edge_pos)

    if not isinstance(edge_color, str) \
           and np.iterable(edge_color) \
           and len(edge_color) == len(edge_vertices):
        if np.alltrue([isinstance(c, str)
                         for c in edge_color]):
            # (should check ALL elements)
            # list of color letters such as ['k','r','k',...]
            edge_colors = tuple([colorConverter.to_rgba(c, alpha)
                                 for c in edge_color])
        elif np.alltrue([not isinstance(c, str)
                           for c in edge_color]):
            # If color specs are given as (rgb) or (rgba) tuples, we're OK
            if np.alltrue([cb.iterable(c) and len(c) in (3, 4)
                             for c in edge_color]):
                edge_colors = tuple(edge_color)
            else:
                # numbers (which are going to be mapped with a colormap)
                edge_colors = None
        else:
            raise ValueError('edge_color must consist of either color names or numbers')
    else:
        if isinstance(edge_color, str) or len(edge_color) == 1:
            edge_colors = (colorConverter.to_rgba(edge_color, alpha), )
        else:
            raise ValueError('edge_color must be a single color or list of exactly m colors where m is the number or edges')

    if tapered:
        edge_collection = PolyCollection(edge_vertices,
                                         facecolors=edge_colors,
                                         linewidths=0,
                                         antialiaseds=(1,),
                                         transOffset = ax.transData,
                                         )
    else:
        edge_collection = LineCollection(edge_pos,
                                         colors=edge_colors,
                                         linewidths=lw,
                                         antialiaseds=(1,),
                                         linestyle=style,
                                         transOffset = ax.transData,
                                         )

    edge_collection.set_zorder(1)  # edges go behind nodes
    edge_collection.set_label(label)
    ax.add_collection(edge_collection)

    # Set alpha globally if provided as a scalar.
    if isinstance(alpha, numbers.Number):
        edge_collection.set_alpha(alpha)
    if edge_colors is None:
        if edge_cmap is not None:
            assert(isinstance(edge_cmap, Colormap))
        edge_collection.set_array(np.asarray(edge_color))
        edge_collection.set_cmap(edge_cmap)
        if edge_vmin is not None or edge_vmax is not None:
            edge_collection.set_clim(edge_vmin, edge_vmax)
        else:
            edge_collection.autoscale()

    # update view
    minx = np.amin(np.ravel(edge_vertices[:, :, 0]))
    maxx = np.amax(np.ravel(edge_vertices[:, :, 0]))
    miny = np.amin(np.ravel(edge_vertices[:, :, 1]))
    maxy = np.amax(np.ravel(edge_vertices[:, :, 1]))

    w = maxx-minx
    h = maxy-miny
    padx,  pady = 0.05*w, 0.05*h
    corners = (minx-padx, miny-pady), (maxx+padx, maxy+pady)
    ax.update_datalim(corners)
    ax.autoscale_view()

    return edge_collection

def draw_nx_nodes(G, pos,
                  nodelist=None,
                  node_size=300,
                  node_color='r',
                  node_shape='o',
                  highlight=None,
                  cmap=None,
                  vmin=None,
                  vmax=None,
                  ax=None,
                  linewidths=0.5,
                  label=None,
                  **kwds):
    """Draw the nodes of the graph G.
    This draws only the nodes of the graph G.
    Parameters
    ----------
    G : graph
       A networkx graph
    pos : dictionary
       A dictionary with nodes as keys and positions as values.
       Positions should be sequences of length 2.
    ax : Matplotlib Axes object, optional
       Draw the graph in the specified Matplotlib axes.
    nodelist : list, optional
       Draw only specified nodes (default G.nodes())
    node_size : scalar or array
       Size of nodes (default=300).  If an array is specified it must be the
       same length as nodelist.
    node_color : color string, or array of floats
       Node color. Can be a single color format string (default='r'),
       or a  sequence of colors with the same length as nodelist.
       If numeric values are specified they will be mapped to
       colors using the cmap and vmin,vmax parameters.  See
       matplotlib.scatter for more details.
    node_shape :  string
       The shape of the node.  Specification is as matplotlib.scatter
       marker, one of 'so^>v<dph8' (default='o').
    highlight : array of boolean, if true a black ring is draw round the node
    cmap : Matplotlib colormap
       Colormap for mapping intensities of nodes (default=None)
    vmin,vmax : floats
       Minimum and maximum for node colormap scaling (default=None)
    linewidths : [None | scalar | sequence]
       Line width of symbol border (default =1.0)
    label : [None| string]
       Label for legend
    Returns
    -------
    matplotlib.collections.PathCollection
        `PathCollection` of the nodes.
    Examples
    --------
    >>> G=nx.dodecahedral_graph()
    >>> nodes=nx.draw_networkx_nodes(G,pos=nx.spring_layout(G))
    Also see the NetworkX drawing examples at
    http://networkx.github.io/documentation/latest/gallery.html
    See Also
    --------
    draw()
    draw_networkx()
    draw_networkx_edges()
    draw_networkx_labels()
    draw_networkx_edge_labels()
    """
    if ax is None:
        ax = plt.gca()

    if nodelist is None:
        nodelist = list(G)

    if not nodelist or len(nodelist) == 0:  # empty nodelist, no drawing
        return None

    try:
        xy = np.asarray([pos[v] for v in nodelist])
    except KeyError as e:
        raise nx.NetworkXError('Node %s has no position.' % e)
    except ValueError:
        raise nx.NetworkXError('Bad value in node positions.')

    node_collection = ax.scatter(xy[:, 0], xy[:, 1],
                                 s=node_size,
                                 c=node_color,
                                 edgecolors='gray',
                                 marker=node_shape,
                                 cmap=cmap,
                                 vmin=vmin,
                                 vmax=vmax,
                                 alpha=1.0,
                                 linewidths=linewidths,
                                 label=label)
    node_collection.set_zorder(2)
    if highlight is not None and len(highlight) == len(xy):
        hxy = xy[highlight]
        hsz = node_size[highlight]
        highlight_collection = ax.scatter(hxy[:, 0], hxy[:, 1],
                                 s=20+2*hsz,
                                 c='none',
                                 edgecolors='k',
                                 marker='o')
        highlight_collection.set_zorder(2)

def getCategoricalColors(nw, df, node_attr, node_vals, color_edges, colors, edge_color='gray'):
    if colors is None:
        colors = ["red", "limegreen", "orange", "mediumblue", "yellow", "darkviolet",
                  "darkred", "darkgreen", "chocolate", "dodgerblue", "gold", "rebeccapurple",
                  "tomato", "seagreen", "navajowhite", "lightskyblue", "lemonchiffon", "plum"]
    counts = df[node_attr].value_counts()
    # color nodes by attribute
    if node_vals is not None:
        #attr_colors = {val: colors[idx%len(colors)] else 'gray' for idx, val in enumerate(node_vals)}
        attr_colors = {val: colors[idx] if idx < len(colors) else 'gray' for idx, val in enumerate(node_vals)}
    else:
        #attr_colors = {val: colors[idx%len(colors)] for idx, val in enumerate(counts.index)}
        attr_colors = {val: colors[idx] if idx < len(colors) else 'gray' for idx, val in enumerate(counts.index)}
    node_color_map = dict(zip(df['id'], df[node_attr].map(attr_colors)))
    node_colors = [node_color_map[node] for node in nw.nodes()]
    # color edges
    if color_edges:
        edge_colors = [node_color_map[edge[0]] if node_color_map[edge[0]] == node_color_map[edge[1]] else edge_color for edge in nw.edges()]
    else:
        edge_colors = [edge_color]*nw.number_of_edges()
    return np.array(node_colors), np.array(edge_colors), attr_colors

# draw network where nodes are colored by a categorical attribute node_attr
# nodes are ordered by descending frequency of category values
# optionally color_edges by whether they are intra (colored) or intra (gray) category
# optional list of node_attr values allows drawing network with predefined attribute value ordering
# so multiple images can have the same color scheme
# if plotfile is given, image is written to file
def draw_network_categorical(nw, df, node_attr='Cluster', node_vals=None, node_size=30,
                             plotfile=None, colors=None, title=None, color_edges=True, draw_edges=True):
    layout = dict(zip(df.id, list(zip(df.x, df.y))))
    node_colors, edge_colors, attr_colors = getCategoricalColors(nw, df, node_attr, node_vals, color_edges, None)
    # plot network
    fig = plt.figure(figsize=(10,8), tight_layout={'rect': (0, 0, 0.8, 1)})
    plt.axis('off')
    _draw_networkx_(nw, layout, arrow=False,
                     with_labels=False, node_color=node_colors,
                     edge_color=edge_colors, node_size=node_size, draw_edges=draw_edges)
    if title is not None:
        plt.gca().set_title(title)
    # add legend
    counts = df[node_attr].value_counts()
    patches = [mpatches.Patch(color=attr_colors[val], label=val+'('+str(counts.loc[val])+')') for val in counts.index]
    fig.axes[0].legend(handles=patches,loc='upper right', bbox_to_anchor=(1.25, 1.0), borderpad=0.05, borderaxespad=0.05)
    if plotfile is not None:
        plt.savefig(plotfile)




# -*- coding: utf-8 -*-
#
# build similarity network from Web of Science data

import os

from WOSTags.MergeWOSFiles import mergeWOSFiles
from WOSTags.ProcessReferences import processRawWoSData
from WOSTags.ProcessReferences import loadAndFilterDocuments
from Tags.BuildKeywords import buildKeywords
from Network.BuildNetwork import buildTagNetwork

# columns to delete in the output
dropCols = ['text', 'AuKeywords', 'KeywordsPlus', 'keywords']

namebase = "Example"         # dataset/project name
rawfilebase = "savedrecs"   # name base of raw, unjoined WoS data files
outfile = namebase+"Raw.txt"
basepath = "Data"
datapath = os.path.join(basepath, namebase)

# concatentate multiple WOS data files
mergeWOSFiles(datapath, rawfilebase, outfile)
# extract smaller set of useful data from raw WOS output
processRawWoSData(datapath, namebase)

# if desired, make keyword blacklist and whitelist
blacklist = set([])
whitelist = set([])

fname = os.path.join(datapath, namebase+"Final.txt")

# too many documents - keep a fraction of them by year and citation rate
#df = loadAndFilterDocuments(fname, 0.9)
df = loadAndFilterDocuments(fname, 0.1)    # for debugging so it runs faster

# set up output file names
nwname = os.path.join(datapath, namebase+".xlsx")
nodesname = os.path.join(datapath, namebase+"Nodes.csv")
edgesname = os.path.join(datapath, namebase+"Edges.csv")
plotname = os.path.join(datapath, namebase+"Plot.pdf")

# build and enhance the keywords, add to df
kwAttr = buildKeywords(df, blacklist, whitelist)


# build network linked by keyword similarity
buildTagNetwork(df, kwAttr, dropCols=dropCols, outname=nwname,
                    nodesname=nodesname, edgesname=edgesname, plotfile=plotname,
                    doLayout=True, clusteredLayout=True)#, draw=True)



# -*- coding: utf-8 -*-
"""
This module implements Louvain community detection.

Modified by Rich Williams to compute using Directed Modularity on directed graphs

"""
#    Copyright (C) 2009 by
#    Thomas Aynaud <thomas.aynaud@lip6.fr>
#    All rights reserved.
#    BSD license.

__PASS_MAX = -1
__MIN = 0.0000001

import networkx as nx
import array

def partition_at_level(dendrogram, level) :
    """Return the partition of the nodes at the given level

    A dendrogram is a tree and each level is a partition of the graph nodes.
    Level 0 is the first partition, which contains the smallest communities, and the best is len(dendrogram) - 1.
    The higher the level is, the bigger are the communities

    Parameters
    ----------
    dendrogram : list of dict
       a list of partitions, ie dictionnaries where keys of the i+1 are the values of the i.
    level : int
       the level which belongs to [0..len(dendrogram)-1]

    Returns
    -------
    partition : dictionnary
       A dictionary where keys are the nodes and the values are the set it belongs to

    Raises
    ------
    KeyError
       If the dendrogram is not well formed or the level is too high

    See Also
    --------
    best_partition which directly combines partition_at_level and generate_dendrogram to obtain the partition of highest modularity

    Examples
    --------
    >>> G=nx.erdos_renyi_graph(100, 0.01)
    >>> dendo = generate_dendrogram(G)
    >>> for level in range(len(dendo) - 1) :
    >>>     print "partition at level", level, "is", partition_at_level(dendo, level)
    """
    partition = dendrogram[0].copy()
    for index in range(1, level + 1) :
        for node, community in partition.items() :
            partition[node] = dendrogram[index][community]
    return partition


'''
def modularity(partition, graph) :
    """Compute the modularity of a partition of a graph

    Parameters
    ----------
    partition : dict
       the partition of the nodes, i.e a dictionary where keys are their nodes and values the communities
    graph : networkx.Graph
       the networkx graph which is decomposed

    Returns
    -------
    modularity : float
       The modularity

    Raises
    ------
    KeyError
       If the partition is not a partition of all graph nodes
    ValueError
        If the graph has no link
    TypeError
        If graph is not a networkx.Graph

    References
    ----------
    .. 1. Newman, M.E.J. & Girvan, M. Finding and evaluating community structure in networks. Physical Review E 69, 26113(2004).

    Examples
    --------
    >>> G=nx.erdos_renyi_graph(100, 0.01)
    >>> part = best_partition(G)
    >>> modularity(part, G)
    """
    if type(graph) != nx.Graph :
        raise TypeError("Bad graph type, use only non directed graph")

    inc = dict([])
    deg = dict([])
    links = graph.size(weight='weight')
    if links == 0 :
        raise ValueError("A graph without link has an undefined modularity")

    for node in graph :
        com = partition[node]
        deg[com] = deg.get(com, 0.) + graph.degree(node, weight = 'weight')
        for neighbor, datas in graph[node].items() :
            weight = datas.get("weight", 1)
            if partition[neighbor] == com :
                if neighbor == node :
                    inc[com] = inc.get(com, 0.) + float(weight)
                else :
                    inc[com] = inc.get(com, 0.) + float(weight) / 2.

    res = 0.
    for com in set(partition.values()) :
        res += (inc.get(com, 0.) / links) - (deg.get(com, 0.) / (2.*links))**2
    return res
'''

def best_partition(graph, partition = None) :
    """Compute the partition of the graph nodes which maximises the modularity
    (or try..) using the Louvain heuristices

    This is the partition of highest modularity, i.e. the highest partition of the dendrogram
    generated by the Louvain algorithm.

    Parameters
    ----------
    graph : networkx.Graph
       the networkx graph which is decomposed
    partition : dict, optionnal
       the algorithm will start using this partition of the nodes. It's a dictionary where keys are their nodes and values the communities

    Returns
    -------
    partition : dictionnary
       The partition, with communities numbered from 0 to number of communities

    Raises
    ------
    NetworkXError
       If the graph is not Eulerian.

    See Also
    --------
    generate_dendrogram to obtain all the decompositions levels

    Notes
    -----
    Uses Louvain algorithm

    References
    ----------
    .. 1. Blondel, V.D. et al. Fast unfolding of communities in large networks. J. Stat. Mech 10008, 1-12(2008).

    Examples
    --------
    >>>  #Basic usage
    >>> G=nx.erdos_renyi_graph(100, 0.01)
    >>> part = best_partition(G)

    >>> #other example to display a graph with its community :
    >>> #better with karate_graph() as defined in networkx examples
    >>> #erdos renyi don't have true community structure
    >>> G = nx.erdos_renyi_graph(30, 0.05)
    >>> #first compute the best partition
    >>> partition = community.best_partition(G)
    >>>  #drawing
    >>> size = float(len(set(partition.values())))
    >>> pos = nx.spring_layout(G)
    >>> count = 0.
    >>> for com in set(partition.values()) :
    >>>     count = count + 1.
    >>>     list_nodes = [nodes for nodes in partition.keys()
    >>>                                 if partition[nodes] == com]
    >>>     nx.draw_networkx_nodes(G, pos, list_nodes, node_size = 20,
                                    node_color = str(count / size))
    >>> nx.draw_networkx_edges(G,pos, alpha=0.5)
    >>> plt.show()
    """
    dendo = generate_dendrogram(graph, partition)
    return partition_at_level(dendo, len(dendo) - 1 )


def generate_dendrogram(graph, part_init = None) :
    """Find communities in the graph and return the associated dendrogram

    A dendrogram is a tree and each level is a partition of the graph nodes.  Level 0 is the first partition, which contains the smallest communities, and the best is len(dendrogram) - 1. The higher the level is, the bigger are the communities


    Parameters
    ----------
    graph : networkx.Graph
        the networkx graph which will be decomposed
    part_init : dict, optionnal
        the algorithm will start using this partition of the nodes. It's a dictionary where keys are their nodes and values the communities

    Returns
    -------
    dendrogram : list of dictionaries
        a list of partitions, ie dictionnaries where keys of the i+1 are the values of the i. and where keys of the first are the nodes of graph

    Raises
    ------
    TypeError
        If the graph is not a networkx.Graph

    See Also
    --------
    best_partition

    Notes
    -----
    Uses Louvain algorithm

    References
    ----------
    .. 1. Blondel, V.D. et al. Fast unfolding of communities in large networks. J. Stat. Mech 10008, 1-12(2008).

    Examples
    --------
    >>> G=nx.erdos_renyi_graph(100, 0.01)
    >>> dendo = generate_dendrogram(G)
    >>> for level in range(len(dendo) - 1) :
    >>>     print "partition at level", level, "is", partition_at_level(dendo, level)
    """
    if type(graph) != nx.Graph and type(graph) != nx.DiGraph:
        raise TypeError("Bad graph type, cannot use on MultiGraph")

    #special case, when there is no link
    #the best partition is everyone in its community
    if graph.number_of_edges() == 0 :
        part = dict([])
        for node in graph.nodes() :
            part[node] = node
        return part

    current_graph = graph.copy()
    status = Status()
    status.init(current_graph, part_init)
    mod = __modularity(status)
    status_list = list()
    __one_level(current_graph, status)
    new_mod = __modularity(status)
    partition = __renumber(status.node2com)
    status_list.append(partition)
    mod = new_mod
    current_graph = induced_graph(partition, current_graph)
    status.init(current_graph)

    while True :
        __one_level(current_graph, status)
        new_mod = __modularity(status)
        if new_mod - mod < __MIN :
            break
        partition = __renumber(status.node2com)
        status_list.append(partition)
        mod = new_mod
        current_graph = induced_graph(partition, current_graph)
        status.init(current_graph)
    return status_list[:]


def induced_graph(partition, graph) :
    """Produce the graph where nodes are the communities

    there is a link of weight w between communities if the sum of the weights of the links between their elements is w

    Parameters
    ----------
    partition : dict
       a dictionary where keys are graph nodes and  values the part the node belongs to
    graph : networkx.Graph
        the initial graph

    Returns
    -------
    g : networkx.Graph
       a networkx graph where nodes are the parts

    Examples
    --------
    >>> n = 5
    >>> g = nx.complete_graph(2*n)
    >>> part = dict([])
    >>> for node in g.nodes() :
    >>>     part[node] = node % 2
    >>> ind = induced_graph(part, g)
    >>> goal = nx.Graph()
    >>> goal.add_weighted_edges_from([(0,1,n*n),(0,0,n*(n-1)/2), (1, 1, n*(n-1)/2)])
    >>> nx.is_isomorphic(int, goal)
    True
    """
    ret = nx.Graph()
    ret.add_nodes_from(partition.values())

    for node1, node2, datas in graph.edges(data = True) :
        weight = datas.get("weight", 1)
        com1 = partition[node1]
        com2 = partition[node2]
        w_prec = ret.get_edge_data(com1, com2, {"weight":0}).get("weight", 1)
        ret.add_edge(com1, com2, weight = w_prec + weight)

    return ret


def __renumber(dictionary) :
    """Renumber the values of the dictionary from 0 to n
    """
    count = 0
    ret = dictionary.copy()
    new_values = dict([])

    for key in dictionary.keys() :
        value = dictionary[key]
        new_value = new_values.get(value, -1)
        if new_value == -1 :
            new_values[value] = count
            new_value = count
            count = count + 1
        ret[key] = new_value

    return ret


def __load_binary(data) :
    """Load binary graph as used by the cpp implementation of this algorithm
    """
    data = open(data, "rb")

    reader = array.array("I")
    reader.fromfile(data, 1)
    num_nodes = reader.pop()
    reader = array.array("I")
    reader.fromfile(data, num_nodes)
    cum_deg = reader.tolist()
    num_links = reader.pop()
    reader = array.array("I")
    reader.fromfile(data, num_links)
    links = reader.tolist()
    graph = nx.Graph()
    graph.add_nodes_from(range(num_nodes))
    prec_deg = 0

    for index in range(num_nodes) :
        last_deg = cum_deg[index]
        neighbors = links[prec_deg:last_deg]
        graph.add_edges_from([(index, int(neigh)) for neigh in neighbors])
        prec_deg = last_deg

    return graph


def __one_level(graph, status) :
    """Compute one level of communities
    """
    modif = True
    nb_pass_done = 0
    cur_mod = __modularity(status)
    new_mod = cur_mod

    while modif  and nb_pass_done != __PASS_MAX :
        cur_mod = new_mod
        modif = False
        nb_pass_done += 1

        for node in graph.nodes() :
            com_node = status.node2com[node]
            degc_totw = status.gdegrees.get(node, 0.) / (status.total_weight*2.)
            neigh_communities = __neighcom(node, graph, status)
            __remove(node, com_node,
                    neigh_communities.get(com_node, 0.), status)
            best_com = com_node
            best_increase = 0
            for com, dnc in neigh_communities.items() :
                incr =  dnc  - status.degrees.get(com, 0.) * degc_totw
                if incr > best_increase :
                    best_increase = incr
                    best_com = com
            __insert(node, best_com,
                    neigh_communities.get(best_com, 0.), status)
            if best_com != com_node :
                modif = True
        new_mod = __modularity(status)
        if new_mod - cur_mod < __MIN :
            break


class Status :
    """
    To handle several data in one struct.

    Could be replaced by named tuple, but don't want to depend on python 2.6
    """
    node2com = {}
    total_weight = 0
    internals = {}
    degrees = {}
    indegrees = {}
    outdegrees = {}
    gdegrees = {}
    directed = False

    def __init__(self) :
        self.node2com = {}
        self.total_weight = 0
        self.degrees = {}
        self.inoutdegrees = {}
        self.gdegrees = {}
        self.gindegrees = {}
        self.goutdegrees = {}
        self.internals = {}
        self.loops = {}
        self.directed = False

    def __str__(self) :
        return ("node2com : " + str(self.node2com) + " degrees : "
            + str(self.degrees) + " internals : " + str(self.internals)
            + " total_weight : " + str(self.total_weight))

    def copy(self) :
        """Perform a deep copy of status"""
        new_status = Status()
        new_status.node2com = self.node2com.copy()
        new_status.internals = self.internals.copy()
        new_status.degrees = self.degrees.copy()
        new_status.inoutdegrees = self.inoutdegrees.copy()
        new_status.gdegrees = self.gdegrees.copy()
        new_status.gindegrees = self.gindegrees.copy()
        new_status.goutdegrees = self.goutdegrees.copy()
        new_status.total_weight = self.total_weight
        new_status.directed = self.directed

    def init(self, graph, part = None) :
        """Initialize the status of a graph with every node in one community"""
        count = 0
        self.node2com = {}
        self.total_weight = 0
        self.degrees = {}
        self.inoutdegrees = {}
        self.gdegrees = {}
        self.gindegrees = {}
        self.goutdegrees = {}
        self.internals = {}
        self.total_weight = graph.size(weight = 'weight')
        self.directed = (type(graph) == nx.DiGraph)
        if part == None :
            for node in graph.nodes() :
                self.node2com[node] = count
                deg = float(graph.degree(node, weight = 'weight'))
                if deg < 0 :
                    raise ValueError("Bad graph type, use positive weights")
                self.degrees[count] = deg
                self.gdegrees[node] = deg
                if self.directed:
                    indeg = float(graph.in_degree(node, weight = 'weight'))
                    outdeg = float(graph.out_degree(node, weight = 'weight'))
                    self.gindegrees[node] = indeg
                    self.goutdegrees[node] = outdeg
                    self.inoutdegrees[count] = indeg*outdeg
                self.loops[node] = float(graph.get_edge_data(node, node,
                                                 {"weight":0}).get("weight", 1))
                self.internals[count] = self.loops[node]
                count = count + 1
        else :
            for node in graph.nodes() :
                com = part[node]
                self.node2com[node] = com
                deg = float(graph.degree(node, weight = 'weight'))
                self.degrees[com] = self.degrees.get(com, 0) + deg
                self.gdegrees[node] = deg
                if self.directed:
                    indeg = float(graph.in_degree(node, weight = 'weight'))
                    outdeg = float(graph.out_degree(node, weight = 'weight'))
                    self.gindegrees[node] = indeg
                    self.goutdegrees[node] = outdeg
                    self.inoutdegrees[count] = indeg*outdeg
                inc = 0.
                inout = 0.
                for neighbor, datas in graph[node].items() :
                    weight = float(datas.get("weight", 1))
                    if weight <= 0 :
                        raise ValueError("Bad graph type, use positive weights")
                    if part[neighbor] == com :
                        if neighbor == node :
                            inc += weight
                        else :
                            inc += weight / 2.
                        if self.directed:
                            inout += indeg*weight
                self.internals[com] = self.internals.get(com, 0) + inc
                if self.directed:
                    self.inoutdegrees[com] = self.inoutdegrees.get(com, 0) + inout



def __neighcom(node, graph, status) :
    """
    Compute the communities in the neighborood of node in the graph given
    with the decomposition node2com
    """
    weights = {}
    for neighbor, datas in graph[node].items() :
        if neighbor != node :
            weight = datas.get("weight", 1)
            neighborcom = status.node2com[neighbor]
            weights[neighborcom] = weights.get(neighborcom, 0) + weight

    return weights


def __remove(node, com, weight, status) :
    """ Remove node from community com and modify status"""
    status.degrees[com] = ( status.degrees.get(com, 0.)
                                    - status.gdegrees.get(node, 0.) )
    status.internals[com] = float( status.internals.get(com, 0.) -
                weight - status.loops.get(node, 0.) )
    status.node2com[node] = -1
    if status.directed:
        status.inoutdegrees[com] = ( status.inoutdegrees.get(com, 0.)
                                    - status.gindegrees.get(node, 0.)*status.goutdegrees.get(node, 0.) )



def __insert(node, com, weight, status) :
    """ Insert node into community and modify status"""
    status.node2com[node] = com
    status.degrees[com] = ( status.degrees.get(com, 0.) +
                                status.gdegrees.get(node, 0.) )
    status.internals[com] = float( status.internals.get(com, 0.) +
                        weight + status.loops.get(node, 0.) )
    if status.directed:
        status.inoutdegrees[com] = ( status.inoutdegrees.get(com, 0.)
                                    + status.gindegrees.get(node, 0.)*status.goutdegrees.get(node, 0.) )


def __modularity(status) :
    """
    Compute the modularity of the partition of the graph quickly using status precomputed
    """
    links = float(status.total_weight)
    result = 0.
    for community in set(status.node2com.values()) :
        in_degree = status.internals.get(community, 0.)
        if status.directed:
            inoutdegree = status.degrees.get(community, 0.)
            if links > 0:
                result += in_degree/links - inoutdegree/(links**2)
        else:
            degree = status.degrees.get(community, 0.)
            if links > 0 :
                result += in_degree / links - ((degree / (2.*links))**2)
    return result



# -*- coding: utf-8 -*-

import math
import numpy as np
import networkx as nx
import scipy.sparse as sps
from sklearn.manifold import TSNE


def setup_layout_distances(nw, nodesdf, dists, maxdist, cluster):
    def computeShortPathUpToMax(adj, maxlen):
        dist = np.empty(adj.shape)
        dist.fill(-1)   # initialize to -1 - any element with -1 hasn't been filled yet
        np.fill_diagonal(dist, 0)
        cc = sps.identity(dist.shape[0])
        for idx in range(1, maxlen + 1):
            cc = cc.dot(adj)
            dist[(dist == -1) & (cc > 0).toarray()] = math.log(1 + idx)   # fill as yet unfilled path lengths
        dist[dist == -1] = 2.0 * maxlen
        return dist

    # compute network shortest path lengths up to a maximum threshold - beyond that things are far away
    if dists is None:
        print("Computing shortest paths")
        dists = computeShortPathUpToMax(nx.adjacency_matrix(nw), maxdist)
    # reduce path length if the nodes are in the same cluster
    clus = nodesdf.Cluster.copy().sort_index().to_numpy()
    if nodesdf is not None and cluster in nodesdf:
        adj = nx.adjacency_matrix(nw).todense()
        for edge in nw.edges:
            adj[edge] = clus[edge[0]] == clus[edge[1]]
        dists = dists - (adj / 1.5)
    return dists, clus


# run tSNE to layout the nodes in 2D space
# dist is a distance matrix.  If None, distances are computed using shortest paths
# paths longer then maxdist are assumed to be "long" and set to 2*maxdist
# returns dict of {nodeid: [x,y]}
# offset increases minimum and so decreases relative distance between nodes, to hopefully spread tight clusters
def runTSNELayout(nw, nodesdf=None, dists=None, maxdist=5, cluster=None):
    # adj is adjacency matrix
    # compute shortest paths up to a max path length
    # fill all longer paths with twice the max
    def initial_positions(all_clus):
        r = 10
        clus = np.unique(all_clus)
        n_clus = len(clus)
        d_phi = 2 * np.pi / n_clus
        clus_idx = dict(zip(clus, np.arange(n_clus)))
        phi = np.array([d_phi * clus_idx[cl] for cl in all_clus])
        pos = r * np.random.rand(len(all_clus))
        x_pos = pos * np.cos(phi)
        y_pos = pos * np.sin(phi)
        return np.stack([x_pos, y_pos]).T

    print("Running tSNE layout")
    dists, clus = setup_layout_dists(nw, nodesdf, dists, maxdist, cluster)
    perp = min(50, nw.number_of_nodes()/10)
    # compute tSNE
    print("Computing tSNE")
    layout = TSNE(n_components=2, metric='precomputed', init=initial_positions(clus),
                  early_exaggeration=5, perplexity=perp).fit_transform(dists)
    # build the output data structure
    nodes = nw.nodes()
    nodeMap = dict(zip(nodes, range(len(nodes))))
    layout_dict = {node: layout[nodeMap[node]] for node in nodes}
    print("Done running tSNE layout")
    # return both networkx style dict and array of positions
    return layout_dict, layout




import umap
from Network.tSNELayout import setup_layout_dists


def runUMAPlayout(nw, nodesdf=None, dists=None, maxdist=5, cluster=None):
    print("Running UMAP layout")
    dists, clus = setup_layout_dists(nw, nodesdf, dists, maxdist, cluster)
    model = umap.UMAP(metric='precomputed')
    layout = model.fit_transform(dists)
    # build the output data structure
    nodes = nw.nodes()
    nodeMap = dict(zip(nodes, range(len(nodes))))
    layout_dict = {node: layout[nodeMap[node]] for node in nodes}
    print("Done running UMAP layout")
    # return both networkx style dict and array of positions
    return layout_dict, layout



# -*- coding: utf-8 -*-

import re
from collections import Counter

# build keywords dataset with keywords and a block of text
# blacklist and whitelist are both sets
# kwAttr is name of keyword column
# keyword column is a string of | separated keywords
# for each document, split text into 1-, 2-, 3-grams using stopwords and punctuation to separate phrases
# add each ngram to keyword list if it is present in the master keyword list
# the master keyword list is all keywords from kwAttr plus the whitelist
# "enhance" # splits multi-word keywords and adds sub-keywords if in the master keyword list
# syndic is a synonym dictionary {synonym:commonTerm} pairs
# all_text == True keeps all text-derived ngrams instead of matching to master list
def buildKeywords(df, blacklist, whitelist, replace_with_space={}, kwAttr='keywords', txtAttr='text', 
                  syndic=None, addFromText=True, enhance=True, all_text=False, include_unigrams=False):
    def addTextKeywords(df, allKwds, all_text):
        stopwords = set(['a', 'an', 'the', 'this', 'that', 'and', 'or', 'of', 'not', 'at',
                         'is', 'in', 'it', 'its', 'but', 'what', 'with', 'as', 'to',
                         'why', 'are', 'do', 'from', 'for', 'on'])

        def findKeywords(ngrams):
            if all_text:
                return ngrams
            textkwds = set()
            for ngram in ngrams:
                if ngram in allKwds:
                    textkwds.add(ngram)
                #else:
                   #print("keyword not found")
            return list(textkwds)

        def add23Grams(wordlist):
            #print ("making bigrams and trigrams from text ngrams")
            bigrams = [' '.join(wordlist[i:i+2]) for i in range(len(wordlist)-1)]
            trigrams = [' '.join(wordlist[i:i+3]) for i in range(len(wordlist)-2)]
            return bigrams+trigrams

        def getNGrams(text):
            wordlists = []
            phrases = re.split('[,.;:!|?]+', text.lower())
            for phrase in phrases:
                words = re.split(' ', phrase)
                currentlist = []
                for w in words:
                    if w in stopwords:
                        if len(currentlist) > 0:
                            wordlists.append(currentlist)
                            currentlist = []
                    else:
                        currentlist.append(w)
                if len(currentlist) > 0:
                    wordlists.append(currentlist)
            ngrams = []
            for wordlist in wordlists:
                if include_unigrams:
                    ngrams.extend(wordlist)  #include all unigrams
                ngrams.extend(add23Grams(wordlist))
            return ngrams

        print ("getting ngrams from text")
        textngrams = df[txtAttr].apply(lambda x: getNGrams(x))  #get 2-, 3-grams out of text
        print ("matching text ngrams to master keyword list to get text keywords")
        textkwds = textngrams.apply(lambda x: findKeywords(x)) #add as keywords if in master list
        newkwds = df.kwds + textkwds  #add ngrams from text if in master kwd list
        newkwds = newkwds.apply(lambda x: (set(x))) # make sure keywords are unique
        return list(newkwds)

    # for each document, split multi-word keywords and for each group of fewer words,
    # add it as a keyword if that keyword appears in the master keyword list
    # reduce n-grams with n > 2 to a set of bigrams
    def enhanceKeywords(df, allKwds):
        def getSubKwd(newKwds, kwdStr, first):
            cnt = len(kwdStr)
            if cnt > 2:
                for i in range(cnt-2):
                    getSubKwd(newKwds, kwdStr[i:cnt-1+i], False)
            if not first:
                kw = ' '.join(kwdStr)
                if kw in allKwds:
                    newKwds.add(' '.join(kwdStr))
        enhancedKwds = []
        for kwds in df['newKwds']:
            newKwds = set()
            for kwd in kwds:
                newKwds.add(kwd)
                getSubKwd(newKwds, kwd.split(), True)
            enhancedKwds.append(list(newKwds))
        return enhancedKwds

    # get a list of all keywords
    kwds = df[kwAttr].str.split('|')
    kwds = kwds.apply(lambda x: [s.strip() for s in x if len(s) > 2])   # only keep kwds with 3 or more characters
    kwds = kwds.apply(lambda x: [s for s in x if s not in blacklist])   # only keep keywords not in blacklist
    kwds = kwds.apply(lambda x: [s.lower() for s in x]) # make all lower case
    if len(replace_with_space) > 0:
        sub_re = '['+ ''.join(replace_with_space)  +']'
        kwds = kwds.apply(lambda x: [re.sub(sub_re, ' ', s) for s in x]) # add whitespace
    kwds = kwds.apply(lambda x: list(set(x))) # make sure keywords are unique
    df['kwds'] = kwds # each value is a list of keywords

    # build keyword histogram for keywords (that occur twice or more)
    kwHist = dict([item for item in Counter([k for kwList in kwds for k in kwList]).most_common() if item[1] > 1])
    print("%d Original Keywords that occur 2 or more times"%len(kwHist))

    if addFromText:
        #add ngrams from title and abstract if they are in the master keyword list
        masterKwds = set(kwHist.keys()).union(whitelist)
        #add search terms (synonyms) and their mapped common term to master list
        if syndic:
            masterKwds.update(syndic.keys(), syndic.values())
        print("adding new keywords to docs if they occur in the text (title+abstract)")
        newkwds = addTextKeywords(df, masterKwds, all_text)
        df['newKwds'] = newkwds
        nKwHist = dict([item for item in Counter([k for kwList in newkwds for k in kwList]).most_common() if item[1] > 1])
        print("%d New Keywords that occur 2 or more times"%len(nKwHist))
    else:
        df['newKwds'] = df['kwds']

    if enhance:
        # split multi-word keywords and add sub-keywords if in the master keyword list
        print("enhancing multi-keywords with sub-keywords")
        df['eKwds'] = enhanceKeywords(df, masterKwds)
    else:
        df['eKwds'] = df['newKwds']
    kwAttr = 'eKwds'
    
    if syndic:
        # synonym dictionary is {term, commonterm}; map terms to common term
        df[kwAttr] = df[kwAttr].apply(lambda x: list(set([syndic[kw] if kw in syndic else kw for kw in x])))

    # recompute histogram on enhanced set of keywords
    kwHist = dict([item for item in Counter([k for kwList in df[kwAttr] for k in kwList]).most_common() if item[1] > 1])
    print("%d Enhanced Keywords that occur 2 or more times"%len(kwHist))
    df['enhanced_keywords'] = df[kwAttr].apply(lambda x: '|'.join(x))
    df.drop(['newKwds', 'kwds'], axis=1, inplace=True)
    return kwAttr



# -*- coding: utf-8 -*-
#
# takes a Web Of Science tab delimited output file and turns it into a tab delimited file ready for text2network
#
# output text: abs - abstract only; all - text+title+keywords; wtd 3x title, 3x keywords 1x text; big - 3x all
#
"""
Created on Wed Mar 26 15:27:02 2014

@author: rich
"""

import os
import csv
import math
import pandas as pd

# extract useful columns, merge text fields, give columns readable names
def processRawWoSData(path, namebase):
    #pubTypes = ['conference', 'book', 'journal', 'book in series', 'patent']

    fieldInfo = {
    'FN':	{'name': 'FileName', 'use': False},
    'VR':	{'name': 'VersionNumber', 'use': False},
    'PT':	{'name': 'PubType', 'use': False},
    'AU':	{'name': 'Authors', 'use': True},
    'AF':	{'name': 'AuthorFullName', 'use': False},
    'CA':	{'name': 'GroupAuthors', 'use': False},
    'TI':	{'name': 'name', 'use': True},
    'ED':	{'name': 'Editors', 'use': True},
    'SO':	{'name': 'PubName', 'use': True},
    'SE':	{'name': 'Book Series Title', 'use': False},
    'BS':	{'name': 'Book Series Subtitle', 'use': False},
    'LA':	{'name': 'Language', 'use': False},
    'DT':	{'name': 'Document Type', 'use': False},
    'CT':	{'name': 'ConferenceTitle', 'use': False},
    'CY':	{'name': 'ConferenceDate', 'use': False},
    'HO':	{'name': 'ConferenceHost', 'use': False},
    'CL':	{'name': 'ConferenceLocation', 'use': False},
    'SP':	{'name': 'ConferenceSponsors', 'use': False},
    'DE':	{'name': 'AuKeywords', 'use': True},
    'ID':	{'name': 'KeywordsPlus', 'use': True},
    'AB':	{'name': 'Abstract', 'use': True},
    'C1':	{'name': 'AuthorAddress', 'use': False},
    'RP':	{'name': 'ReprintAddress', 'use': False},
    'EM':	{'name': 'E-mailAddress', 'use': False},
    'FU':	{'name': 'FunderAndGrant', 'use': False},
    'FX':	{'name': 'FundingText', 'use': False},
    'CR':	{'name': 'Cited References', 'use': False},
    'NR':	{'name': 'ReferenceCount', 'use': True},
    'TC':	{'name': 'TimesCited', 'use': True},
    'PU':	{'name': 'Publisher', 'use': False},
    'PI':	{'name': 'Publisher City', 'use': False},
    'PA':	{'name': 'Publisher Address', 'use': False},
    'SN':	{'name': 'ISSN', 'use': False},
    'BN':	{'name': 'ISBN', 'use': False},
    'J9':	{'name': 'SourceAbbrev', 'use': False},
    'JI':	{'name': 'ISOSourceAbbrev', 'use': False},
    'PD':	{'name': 'PubDate', 'use': False},
    'PY':	{'name': 'PubYear', 'use': True},
    'VL':	{'name': 'Volume', 'use': False},
    'IS':	{'name': 'Issue', 'use': False},
    'PN':	{'name': 'PartNumber', 'use': False},
    'SU':	{'name': 'Supplement', 'use': False},
    'SI':	{'name': 'SpecialIssue', 'use': False},
    'BP':	{'name': 'BeginningPage', 'use': False},
    'EP':	{'name': 'EndingPage', 'use': False},
    'AR':	{'name': 'ArticleNumber', 'use': False},
    'PG':	{'name': 'PageCount', 'use': True},
    'DI':	{'name': 'DOI', 'use': True},
    'SC':	{'name': 'SubjectCategory', 'use': True},
    'GA':	{'name': 'DocDeliveryNumber', 'use': False},
    'UT':	{'name': 'UniqueArticleIdentifier', 'use': False},
    'PM':	{'name': 'PubMedID', 'use': False},
    'ER':	{'name': 'EndRecord', 'use': False},
    'EF':	{'name': 'EndFile', 'use': False},
    }

    title = 'TI'
    kwds = 'DE'
    kwds2 = 'ID'
    pubDate = 'PD'
    pubYr = 'PY'
    abstract = 'AB'
    refs = 'CR'
    doi = 'DI'
    sType = 'SearchType'
    sRef = 'SReferences'
    xRef = 'XReferences'
    sxRef = 'SXReferences'
    inbred = 'Inbredness'

    #os.chdir(path)
    dataFile = os.path.join(path, namebase + "Raw.txt")

    # compute reference similarity for each pair of nodes
    def computeReferenceSimilarities(references):
        similarities = {}
        maxSim = 0
        for nodeid, ref in references.items():
            sim = similarities[nodeid] = {}
            refSet = set(ref.split(';'))
            for nodeid2, ref2 in references.items():
                if nodeid < nodeid2:
                    refSet2 = set(ref2.split(';'))
                    nTotal = len(refSet.union(refSet2))
                    nOverlap = len(refSet.intersection(refSet2))
                    if nOverlap > 0 and nTotal > 0:
                        simVal = float(nOverlap)/nTotal
                        sim[nodeid2] = simVal
                        if simVal > maxSim:
                            maxSim = simVal
        return similarities

    # write data, pass in filename and field to use for 'text' in final output
    def writeData(name, fields, data, textField):
        for row in data:
            row['text'] = row[textField]
        f = open(name, 'w')
        writer = csv.DictWriter(f, fields, dialect=csv.excel_tab, extrasaction='ignore')
        writer.writeheader()
        writer.writerows(data)
        f.close()

    # main entry point for processing results from web of science searches.
    # hasSearchType is true if WoS results have an added searchType column,
    # used to combine and differentiate results from multiple searches
    def processWoSData(dataFile, hasSearchType):
        # get used field names
        fields = []
        for key, value in fieldInfo.items():
            if value['use'] is True:
                fields.append(key)

        # read in and re-arrange the data
        data = []
        doiIndex = {}
        references = {}
        f = open(dataFile, 'rU')
        reader = csv.DictReader(f, dialect=csv.excel_tab)
        #read data one row at a time
        nodeid = 0
        for row in reader:
            # only include records that have abstract text and keywords
            if len(row[abstract]) > 0:
                if nodeid%100 == 0:
                    print("Processed %d"%nodeid)
                kwplus = row[kwds2]
                if len(kwplus) > 0 :
                    kwplus = kwplus.lower()
                dataRow = {}
                # build keywords
                keywords = row[kwds]
                if len(keywords) > 0 and len(kwplus) > 0:
                    keywords += ';'
                keywords += kwplus
                if len(keywords) == 0:
                    continue
                keywords = keywords.lower().replace(';', '|')
                # build text and add to row
                dataRow['alltext'] = row[title] + '. ' + row[abstract] + '. ' + keywords + '.'
                #dataRow['wtdtext'] = row[title] + '. ' + row[title] + '. ' + row[title] + '. ' + row[abstract] + '. ' + keywords + '.' + keywords + '.' + keywords + '.'
                #dataRow['bigtext'] = dataRow['alltext'] + ' ' + dataRow['alltext'] + ' ' + dataRow['alltext']
                # build date string and add to row
                date = ''
                if len(row[pubDate]) > 0:
                    date = row[pubDate] + '-'
                date += row[pubYr]
                dataRow['date'] = date
                dataRow['id'] = nodeid
                dataRow['keywords'] = keywords
                if hasSearchType:
                    dataRow[sType] = row[sType]
                doiIndex[row[doi]] = nodeid
                references[nodeid] = row[refs]
                nodeid += 1
                # copy 'used' fields into row with renamed output fields instead of WoS codes
                for field in fields:
                    info = fieldInfo[field]
                    if field in row:
                        dataRow[info['name']] = row[field]
                data.append(dataRow)
        f.close()

        if hasSearchType:
            # count references to papers with different searchTypes
            for nodeid, ref in references.items():
                nS = 0
                nX = 0
                nSX = 0
                for refItem in ref.split(';'):
                    refFields = refItem.split(',')
                    doiField = refFields[len(refFields) - 1].lstrip()
                    if doiField.startswith("DOI"):
                        refDOI = doiField[4:]   # got a DOI reference
                        if doiIndex.has_key(refDOI):    # reference is in this data set
                            id = doiIndex[refDOI]
                            searchType = data[id][sType]
                            if searchType == 'S':
                                nS += 1
                            elif searchType == 'X':
                                nX += 1
                            else:
                                nSX += 1
                dataRow = data[nodeid]
                dataRow[sRef] = nS
                dataRow[xRef] = nX
                dataRow[sxRef] = nSX
                searchType = dataRow[sType]
                tot = nS+nX
                if tot > 3:     # no significant inbreeding when total number of s or x references is small
                    inbredVal = (2 * nS/tot) - 1        # -1 - all x, +1 - all s
                else:
                    inbredVal = 0
                dataRow[inbred] = inbredVal

        # build array of output fields
        outFields = []
        outFields.append('text')
        outFields.append('date')
        outFields.append('keywords')
        for field in fields:
            outFields.append(fieldInfo[field]['name'])
        if hasSearchType:
            outFields.append(sType)
            outFields.append(sRef)
            outFields.append(xRef)
            outFields.append(sxRef)
            outFields.append(inbred)
        # write out the data
        writeData(os.path.join(path, namebase + "Final.txt"), outFields, data, 'alltext')

    processWoSData(dataFile, False)

# load document file, keep a fraction of documents by year and citation rate
# sort by publication year and citation count
# then iterate through each year keeping documents in each year with most citations
def loadAndFilterDocuments(fname, keepFrac):
    df = pd.read_csv(fname, sep='\t', header=0)

    if keepFrac < 1.0:
        # sort by year and citation count
        df.sort_values(by=['PubYear', 'TimesCited'], inplace=True, ascending=False)
        cnts = df.PubYear.value_counts()
        start = 0
        keep = pd.Series([False]*len(df))
        while start < len(df):
            cnt = cnts[df.PubYear[start]]
            last = int(start + cnt*math.ceil(keepFrac))
            keep[start:last] = True
            start += cnt
        return df[keep.values]
    return df




#
# increment version
# remove egg info, dist
# python setup.py sdist
# python setup.py bdist_wheel --universal
# twine upload dist/*
#

from setuptools import setup
from setuptools import find_packages

setup(name='tag2network',
      version='0.0.8',
      description='Build similarity network from tagged documents',
      long_description="Build similarity network from any dataset where a set of keywords or other \
                          tags is assigned to each document. The network is build by computing \
                          cosine simiarity between each document, based on their tag sets, and \
                          thresholding the similarities to produce a network.  Includes ability to \
                          extract ngrams from blocks of text and use as tags",
      url='https://github.com/foodwebster/tag2network',
      author='foodwebster',
      author_email='foodwebster@gmail.com',
      license='MIT',
      packages=find_packages(),
      install_requires=[
          'numpy', 'scipy', 'pandas', 'networkx'
      ],
      zip_safe=False)



"""
Modified tSNE layout to pull clusters together into visually coherent groups.

1) tSNE on whole network
2) Kamada-Kawai on each cluster
3) Pull 'distant nodes in to limit total radius of each cluster
4) (Optional) GTree to move clsuters to eliminate overlap
"""
#from Network.tSNELayout import runTSNELayout
from tSNELayout import runTSNELayout 
import networkx as nx
import numpy as np
import scipy as sp


def remove_overlap(nodes):
    """Implement GTree algorithm https://arxiv.org/pdf/1608.02653.pdf."""
    nodes = [n.copy() for n in nodes]

    def dist(idx1, idx2, pos, nodes):
        d = pos[idx1] - pos[idx2]
        center_to_center = np.sqrt((d * d).sum())
        return center_to_center - nodes[idx1]['radius'] - nodes[idx2]['radius']

    def get_next(mst, previous, current):
        edges = list(mst.edges(current))
        next_nodes = []
        for e in edges:
            if previous is None or previous not in e:
                next_nodes.append(e[1] if e[0] == current else e[0])
        return next_nodes

    def shift_nodes(nodes, mst, source, target, delta_x, delta_y):
        # shift the target
        trg_node = nodes[target]
        trg_node['x'] += delta_x
        trg_node['y'] += delta_y
        # shift nodes recursively
        next_nodes = get_next(mst, source, target)
        for next_n in next_nodes:
            shift_nodes(nodes, mst, target, next_n, delta_x, delta_y)

    def process_tree(nodes, mst, previous, current):
        # process mst recursively
        next_nodes = get_next(mst, previous, current)
        for next_n in next_nodes:
            wt = mst.edges[(current, next_n)]['weight']
            if wt < 0:
                # compute the shift x, y
                src_node = nodes[current]
                trg_node = nodes[next_n]
                dx = trg_node['x'] - src_node['x']
                dy = trg_node['y'] - src_node['y']
                dist = np.sqrt(dx ** 2 + dy ** 2)
                frac_x = dx / dist
                frac_y = dy / dist
                wt = mst.edges[(current, next_n)]['weight']
                delta_x = -wt * frac_x
                delta_y = -wt * frac_y
                # shift target and its children
                shift_nodes(nodes, mst, current, next_n, delta_x, delta_y)
            process_tree(nodes, mst, current, next_n)

    max_steps = 10
    for step in range(max_steps):
        # extract position data to numpy
        pos = np.array([[n['x'], n['y']] for n in nodes])
        # build delauney triangulation
        tri = sp.spatial.Delaunay(pos)
        # build weighted networkx graph. Weight is distance between node edges
        raw_edges = set()
        for sim in tri.simplices:
            raw_edges.add((sim[0], sim[1]))
            raw_edges.add((sim[1], sim[2]))
            raw_edges.add((sim[2], sim[0]))
        nw = nx.Graph()
        n_overlap = 0
        for e in raw_edges:
            d = dist(e[0], e[1], pos, nodes)
            nw.add_edge(e[0], e[1], weight=d)
            if d < 0:
                n_overlap += 1
        print(f'Step {step} n_overlap = {n_overlap}')
        # quit looping if all weights are positive
        if n_overlap == 0:
            break
        # get minimal spanning tree of weighted graph
        mst = nx.minimum_spanning_tree(nw)
        # roots have degree == 1
        root = [n for n, d in mst.degree if d == 1][0]
        # recursively process mst from root
        process_tree(nodes, mst, None, root)
    return {n['name']: np.array([n['x'], n['y']]) for n in nodes}


def compress_groups(nw, nodes_df, layout_dict, cluster_attr, no_overlap,
                    max_expansion=1.5, scale_factor=1.0):
    clus_nodes = nodes_df.groupby(cluster_attr)
    subgraphs = {clus: cdf.id.to_list() for clus, cdf in clus_nodes}
    new_positions = {}
    clusters = []
    clus_id = 0
    for clus, subg in subgraphs.items():
        # get starting positions (from tSNE layout)
        pos = {k: layout_dict[k] for k in subg}
        clus_pos = np.array(list(pos.values()))
        # get cluster centroid, final scale and distance from center
        center = np.median(clus_pos, axis=0)
        scale = np.sqrt(len(subg)) * scale_factor
        # add cluster node
        clusters.append({'id': clus_id,
                         'name': clus,
                         'x': center[0],
                         'y': center[1],
                         'radius': scale})
        clus_id += 1
    # move cluster centers to remove overlap
    print("Repositioning cluster centers")
    centers = {cl['name']: (cl['x'], cl['y']) for cl in clusters}
    if no_overlap:
        new_centers = remove_overlap(clusters)
    else:
        new_centers = centers
    print("Compressing layout of nodes in clusters")
    for clus, subg in subgraphs.items():
        # get starting positions (from tSNE layout)
        pos = {k: layout_dict[k] for k in subg}
        clus_pos = np.array(list(pos.values()))
        # get cluster centroid, final scale and distance from center
        center = np.median(clus_pos, axis=0)
        scale = np.sqrt(len(subg)) * scale_factor
        dists = np.sqrt(((clus_pos - center) ** 2).sum(axis=1))
        # use a truncated, normalized Mechelis-Menten function
        # to rescale distance from center
        om = max(dists)
        nm = scale
        k = om * nm / (om - nm / 2)
        rescale = np.clip(k / (k / 2 + dists), None, max_expansion)
        center = centers[clus]
        new_center = new_centers[clus]
        new_pos = new_center + ((clus_pos - center) * rescale.reshape(-1, 1))
        new_positions.update(dict(zip(pos.keys(), new_pos)))
    return new_positions


def run_cluster_layout(nw, nodes_df, dists=None, maxdist=5, cluster_attr='Cluster',
                       no_overlap=True, max_expansion=1.5, scale_factor=1.0):
    """
    Pull nodes in clusters in tSNE layout into visually coherent groups.

    1) tSNE on whole network
    2) Kamada-Kawai on each cluster
    3) Pull 'distant nodes in to limit total radius of each cluster
    4) (Optional) GTree to minimally move clsuters to eliminate overlap


    Args:
        nw (TYPE): The network to lay out as a networkx Graph.
        nodes_df (TYPE): The node data in a pandas DataFrame.
        dists (TYPE, optional): Distance matrix, if none path length is used to compute distances.
                                Defaults to None.
        maxdist (TYPE, optional): Maximum path length to consider when computing path-based distances.
                                  Defaults to 5.
        cluster_attr (TYPE, optional): The attribute used to define the clusters. Defaults to 'Cluster'.
        no_overlap (TYPE, optional): If True, do not remove cluster overlap. Defaults to True.
        max_expansion (TYPE, optional): Internode distance expansion at the center of each cluster. Defaults to 1.5.
        scale_factor (TYPE, optional): Cluster area scale factor. Defaults to 1.0.

    Returns
    -------
        new_positions (TYPE): DESCRIPTION.
        layout (TYPE): DESCRIPTION.

    """
    layout_dict, layout = runTSNELayout(nw, nodes_df, dists, maxdist, cluster_attr)
    clus_nodes = nodes_df.groupby(cluster_attr)
    subgraphs = {clus: nw.subgraph(cdf.id.to_list()) for clus, cdf in clus_nodes}
    new_positions = {}
    for clus, subg in subgraphs.items():
        print(f"Laying out subgraph for {clus}")
        # get starting positions (from tSNE layout)
        pos = {k: layout_dict[k] for k in subg.nodes}
        clus_pos = np.array(list(pos.values()))
        # get cluster centroid
        center = np.median(clus_pos, axis=0)
        scale = np.sqrt(len(subg))
        new_pos = nx.kamada_kawai_layout(subg, pos=pos, weight=None,
                                         scale=scale,
                                         center=center)
        new_positions.update(new_pos)
    new_positions = compress_groups(nw, nodes_df, new_positions, cluster_attr,
                                    no_overlap, max_expansion, scale_factor)
    layout = [new_positions[idx] for idx in layout_dict.keys()]
    return new_positions, layout


if __name__ == "__main__":
    # build test dataset
    init_nodes = [
        {'id': 1,
         'name': 'node1',
         'x': -0.5,
         'y': 0,
         'radius': 0.3
         },
        {'id': 2,
         'name': 'node2',
         'x': 0,
         'y': 0,
         'radius': 0.3
         },
        {'id': 3,
         'name': 'node3',
         'x': 0.5,
         'y': 0.1,
         'radius': 0.2
         },
        {'id': 4,
         'name': 'node4',
         'x': 1.0,
         'y': 0.1,
         'radius': 0.4
         }
        ]

    new_centers = remove_overlap(init_nodes)


# -*- coding: utf-8 -*-

import pandas as pd

#from DrawNetwork import draw_network_categorical
from Network.InteractiveNetworkViz import drawInteractiveNW

nodesfile = "tag2network/Data/Example/ExampleNodes.csv"
edgesfile = "tag2network/Data/Example/ExampleEdges.csv"

nodesdf = pd.read_csv(nodesfile)
edgesdf = pd.read_csv(edgesfile)

drawInteractiveNW(nodesdf, edgesdf=edgesdf, color_attr="Cluster", label_attr="name",
                  title="Interactive Network Visualization",
                  plotfile="Data/Example/TestNetwork", inline=False)
